#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAGIC_STRING_LENGTH 12
static const char magic_string[MAGIC_STRING_LENGTH] = "PACKPACKPACK";

#define FIRMWARE_HEADER_LENGTH 4
static const char firmware_header[FIRMWARE_HEADER_LENGTH] = { 0x00, 0x00, 0xa0, 0xe1 };

enum { FIRMWARE_MAGIC_NOT_FOUND, FIRMWARE_HEADER_NOT_FOUND, FIRMWARE_END_NOT_FOUND, FIRMWARE_READ_ERROR, FIRMWARE_WRITE_ERROR, FIRMWARE_SUCCESS };

#define byte_swap(x) (((x & 0xFFFF0000) >> 16) | ((x & 0x0000FFFF) << 16))
static int byteswap;

static int extract_next_firmware(FILE *infile, char *outfilename)
{
  FILE *outfile;
  int i;
  unsigned int b;
  char c;
  
  /* Look for the magic string */
  i = 0;
  while(1) {
    if(fread(&c, 1, 1, infile) != 1) {
      if(feof(infile)) return FIRMWARE_MAGIC_NOT_FOUND; else return FIRMWARE_READ_ERROR;
    }
    if(c == magic_string[i]) {
      i++;
      if(i == MAGIC_STRING_LENGTH) break;
    } else {
      if(c == magic_string[0]) i = 1; else i = 0;
    }
  }
  fseek(infile, -MAGIC_STRING_LENGTH, SEEK_CUR);
  printf("[%s] Magic string found at offset 0x%lx.\n", outfilename, ftell(infile));
  
  /* Look for the beginning of the firmware */
  i = FIRMWARE_HEADER_LENGTH-1;
  while(1) {
    if(fread(&c, 1, 1, infile) != 1) return FIRMWARE_READ_ERROR;
    if(c == firmware_header[i]) {
      i--;
      if(i == -1) break;
    } else {
      if(c == firmware_header[FIRMWARE_HEADER_LENGTH-1]) i = FIRMWARE_HEADER_LENGTH-2; else i = FIRMWARE_HEADER_LENGTH-1;
    }
    if(ftell(infile) == 1) return FIRMWARE_HEADER_NOT_FOUND;
    fseek(infile, -2, SEEK_CUR);
  }
  fseek(infile, -1, SEEK_CUR);
  printf("[%s] Beginning of the firmware found at offset 0x%lx.\n", outfilename, ftell(infile));

  outfile = fopen(outfilename, "wb");
  if(outfile == NULL) return FIRMWARE_WRITE_ERROR;
  /* Detect the end of the firmware (while writing it to the output file) as it ends with its length */
  i = 0;
  while(1) {
    if(fread(&b, 4, 1, infile) != 1) {
      fclose(outfile);
      remove(outfilename);
      if(feof(infile)) return FIRMWARE_END_NOT_FOUND; else return FIRMWARE_READ_ERROR;
    }
    if(fwrite(&b, 4, 1, outfile) != 1) {
      fclose(outfile);
      remove(outfilename);
      return FIRMWARE_WRITE_ERROR;
    }
    if((byteswap ? byte_swap(b) : b) == i) {
      printf("[%s] End of firmware found at offset 0x%lx.\n", outfilename, ftell(infile));
      i += 4;
      break;
    }
    i += 4;
  }
  if(fclose(outfile) != 0) {
    remove(outfilename);
    return FIRMWARE_WRITE_ERROR;
  }
  printf("[%s] Successfully extracted %u bytes of firmware.\n", outfilename, i);
  return FIRMWARE_SUCCESS;
}

int main(int argc, char *argv[])
{
  FILE *fin;
  char *buffer;
  unsigned int firmware_count;
  int i;
 
  puts("fwextract - Prism54 firmware extractor (version 3)");
  if(argc != 4) {
    puts("By Sebastien BOURDEAUDUCQ - sebastien.b@swissinfo.org - http://yoshiyo.ath.cx/seb/");
    puts("Extracts firmware for Prism54-based WLAN cards from the official binary drivers.");
    puts("Usage: fwextract <inputfile> <basename> <byteswap>");
    puts("inputfile\tfile to extract the firmware(s) from");
    puts("basename\tbeginning of the name the extracted firmware(s) will have");
    puts("byteswap\tvalue can be 'byteswap' or 'nobyteswap', depending whether the endiannesses of the host and the driver match");
    return 1;
  }
  if(strcmp(argv[3], "byteswap") == 0) {
    byteswap = 1;
  } else {
    if(strcmp(argv[3], "nobyteswap") == 0) {
      byteswap = 0;
    } else {
      puts("ERROR: please specify 'byteswap' or 'nobyteswap'");
      return 1;
    }
  }
  fin = fopen(argv[1], "rb");
  if(fin == NULL) {
    puts("ERROR: Unable to open input file");
    return 1;
  }
  
  firmware_count = 0;
  while(1) {
    if(asprintf(&buffer,"%s%02d.arm", argv[2], firmware_count) == -1) {
      puts("ERROR: Memory allocation failed.");
      fclose(fin);
      return 1;
    }
    if((i = extract_next_firmware(fin, buffer)) != FIRMWARE_SUCCESS) {
      free(buffer);
      break;
    }
    free(buffer);
    firmware_count++;
  }
  
  fclose(fin);
  
  switch(i) {
    case FIRMWARE_HEADER_NOT_FOUND:
      puts("ERROR: Unable to find ARM NOP.");
      break;
    case FIRMWARE_END_NOT_FOUND:
      puts("ERROR: Unable to find the end of the firmware.");
      break;
    case FIRMWARE_READ_ERROR:
      puts("ERROR: I/O error on input file.");
      break;
    case FIRMWARE_WRITE_ERROR:
      puts("ERROR: I/O error on output file.");
      break;
  }
  if(firmware_count == 0) {
    puts("ERROR: No firmware found.");
    return 1;
  }
  
  return 0;
}
